# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Master solver used in sequential Benders-baded MIQP algorithm."""
import numpy as np
import casadi as ca
from camino.solvers import Stats, MinlpProblem, MinlpData
from camino.utils import colored, toc
from camino.settings import GlobalSettings, Settings
from camino.solvers.decomposition.benders_master import BendersMasterMILP
from camino.solvers.utils import Constraints, get_solutions_pool, any_equal
from enum import Enum
import logging

from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class LowerApproximation:
    """Store info on lower approximation cuts."""

    def __init__(self, x, nu):
        """Store info on lower approximation cuts."""
        self.nr = 0
        self.g = []
        self.dg = []
        self.dg_corrected = []
        self.x_lin = []
        self.multipliers = []
        self.x = x
        self.nu = nu
        self.is_corrected = []

    def add(self, point, offset, gradient, gradient_corrected=None):
        """Add a lower approximation cut."""
        if gradient_corrected is None:
            gradient_corrected = gradient
            self.is_corrected.append(False)
        else:
            self.is_corrected.append(True)

        self.nr += 1
        self.x_lin.append(point)
        self.g.append(offset)
        self.dg.append(gradient)
        self.dg_corrected.append(gradient_corrected)
        if not self.multipliers:
            self.multipliers.append(1)
        else:
            # the new constraint added should have the same multiplier
            # of the others (cf. gradient-amplification strategy)
            self.multipliers.append(self.multipliers[0])

    def __call__(self, x_value, nu=0):
        """Evaluate the bounds."""
        return [
            gi + m * dgi.T @ (x_value - xi) - nu
            for gi, dgi, m, xi in zip(
                self.g, self.dg_corrected, self.multipliers, self.x_lin
            )
        ]

    def to_generic(self, nu=None):
        """Create bounds."""
        if nu is None:
            nu = self.nu
        return Constraints(
            self.nr,
            ca.vertcat(*self(self.x, nu)),
            -ca.inf * np.ones(self.nr),
            np.zeros(self.nr),
        )

    def __add__(self, other):
        """Add bounds."""
        return self.to_generic() + other


def compute_gradient_correction(x_best, x_new, obj_best, obj_new, grad, s: Settings):
    """Compute gradient correction, based on L2 norm."""

    with_ipopt = False
    W = np.eye(x_best.shape[0])  # weighting matrix L2-norm
    W_inv = np.linalg.inv(W)

    if with_ipopt:
        nr_x = x_best.numel()
        grad_corr = GlobalSettings.CASADI_VAR.sym("gradient_correction", nr_x)
        obj = ca.norm_2(grad_corr) ** 2
        g = (obj_new - obj_best - s.EPS) + (grad + grad_corr).T @ (x_best - x_new)
        solver = ca.nlpsol("solver", "ipopt", {"f": obj, "g": g, "x": grad_corr}, {})
        sol = solver(
            x0=np.abs(x_new - x_best),
            lbx=-ca.inf * np.ones(nr_x),
            ubx=ca.inf * np.ones(nr_x),
            lbg=-ca.inf,
            ubg=0,
        )
        print(sol["x"])
        return sol["x"] + grad
    else:
        delta_x = x_best - x_new
        residual = obj_best - obj_new - grad.T @ delta_x
        return grad + (residual / (delta_x.T @ W_inv @ delta_x)) @ W_inv @ delta_x


def compute_infeasibility(g, lbg, ubg):
    """Compute infeasibility using benders-type infeasibility cuts."""
    g = g[: lbg.shape[0]]
    return np.sum(lbg[lbg > g] - g[lbg > g]) + np.sum(g[ubg < g] - ubg[ubg < g])


class BendersRegionMasters(BendersMasterMILP):
    """Mixing the idea from Moritz with a slightly altered version of benders masters."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        with_lb_milp=True,
        with_milp_only=False,
    ):
        """Create the benders constraint MILP."""
        super(BendersRegionMasters, self).__init__(
            problem, data, stats, s, with_lin_bounds=False
        )
        # Settings
        self.alpha_kronqvist = s.ALPHA_KRONQVIST
        self.trust_region_feasibility_rho = s.RHO_AMPLIFICATION
        self.sol_infeasibility = np.inf

        if s.WITH_DEBUG:
            self.problem = problem

        # Setups
        self.setup_common(problem, s)

        self.grad_f_x = ca.Function(
            "gradient_f_x",
            [problem.x, problem.p],
            [ca.gradient(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.jac_g = ca.Function(
            "jac_g",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.f_qp = None
        if problem.f_qp is not None:
            self.f_qp = problem.f_qp

        if problem.gn_hessian is None:
            self.f_hess = ca.Function(
                "hess_f_x",
                [problem.x, problem.p],
                [ca.hessian(problem.f, problem.x)[0]],
            )
        else:
            self.f_hess = ca.Function(
                "hess_f_x", [problem.x, problem.p], [problem.gn_hessian]
            )

        self._x = GlobalSettings.CASADI_VAR.sym("x_benders", problem.x.numel())
        self._x_bin = self._x[problem.idx_x_integer]
        self.idx_g_conv = problem.idx_g_conv

        self.options.update(
            {
                "discrete": [
                    1 if elm in problem.idx_x_integer else 0
                    for elm in range(self._x.shape[0])
                ]
            }
        )
        self.options_master = self.options.copy()
        self.options_master["discrete"] = self.options["discrete"] + [0]
        self.options_master["error_on_fail"] = False
        # self.options_master['gurobi.PoolSolutions'] = 1 # TODO dont force pool 1 but check if pool contains the best solution and in case discard it.
        self.options["error_on_fail"] = False

        self.mipgap_miqp = s.BRMIQP_GAP
        self.mipgap_milp = s.LBMILP_GAP
        if self.settings.MIP_SOLVER == "gurobi":
            self.mipgap_options_str = "gurobi.MIPGap"
            self.mip_timelimit_options_str = "gurobi.TimeLimit"
        elif self.settings.MIP_SOLVER == "highs":
            self.mipgap_options_str = "highs.mip_rel_gap"
            self.mip_timelimit_options_str = "highs.time_limit"
        self.options_master[self.mipgap_options_str] = self.mipgap_milp
        self.with_lb_milp = with_lb_milp
        self.with_milp_only = with_milp_only
        self.hessian_not_psd = problem.hessian_not_psd
        self.with_oa_conv_cuts = True
        self.reset(data)

    def _check_cut_valid(self, g, grad_g, x_best, x_sol, x_sol_obj):
        """Check if the cut is valid."""
        value = g + grad_g.T @ (x_best - x_sol)
        # print(f"Cut valid (lower bound)?: {value} vs real {x_sol_obj}")
        return value - self.settings.EPS <= x_sol_obj

    # Amplifications and corrections
    def clip_gradient(self, current_value, lower_bound, gradients):
        # TODO: lower bound might be wrong when clipping Benders cut
        max_gradient = float(current_value - lower_bound)
        return np.clip(gradients, -max_gradient, max_gradient)

    def _gradient_amplification(self):
        """Amplify the gradient of every new cut with the chosen rho."""
        # Amplification of Benders' cuts
        for i, m in enumerate(self.g_benders.multipliers):
            if self.g_benders.is_corrected[i]:
                if m != self.trust_region_feasibility_rho:
                    self.g_benders.multipliers[i] = self.trust_region_feasibility_rho
        # Amplification of OA objective cuts
        if len(self.g_oa_objective.g) > 0:
            for i, m in enumerate(self.g_oa_objective.multipliers):
                if self.g_oa_objective.is_corrected[i]:
                    if m != self.trust_region_feasibility_rho:
                        self.g_oa_objective.multipliers[i] = (
                            self.trust_region_feasibility_rho
                        )

    def _gradient_corrections_old_cuts(self):
        x_sol_best_bin = self.sol_best["x"][self.idx_x_integer]
        # Check and correct - if necessary - all the points in memory
        for i in range(self.g_benders.nr - 1):
            # New corrections based on previously corrected gradients
            if not self._check_cut_valid(
                self.g_benders.g[i],
                self.g_benders.dg_corrected[i],
                x_sol_best_bin,
                self.g_benders.x_lin[i],
                self.y_N_val,
            ):
                self.g_benders.dg_corrected[i] = compute_gradient_correction(
                    x_sol_best_bin,
                    self.g_benders.x_lin[i],
                    self.y_N_val,
                    self.g_benders.g[i],
                    self.g_benders.dg_corrected[i],
                    self.settings,
                )
                self.g_benders.is_corrected[i] = True
                logger.debug(f"Correcting gradient for lower approx {i}")

    # Infeasibility cuts
    def _add_infeasibility_cut(self, sol, nlpdata):
        """Add infeasibility cut."""
        if "x_infeasible" in sol:
            self._add_infeasible_cut_closest_point(sol)
        else:
            self._add_infeasible_cut(sol["x"], sol["lam_g"], nlpdata)

    def _add_infeasible_cut_closest_point(self, sol):
        """Add infeasible cut closest point."""
        # Rotate currently around x, but preferably around x_infeasible
        dx = (sol["x_infeasible"] - sol["x"])[self.idx_x_integer]
        dx_min = abs(min(dx.full()))
        dx_max = abs(max(dx.full()))
        multiplier = min(1 / max(dx_min, dx_max), 1000)

        if self.sol_best_feasible:
            if not self._check_cut_valid(
                0,
                dx * multiplier,
                self.sol_best["x"][self.idx_x_integer],
                sol["x"][self.idx_x_integer],
                0,
            ):
                grad_corr = compute_gradient_correction(
                    self.sol_best["x"][self.idx_x_integer],
                    sol["x"][self.idx_x_integer],
                    0,
                    0,
                    dx * multiplier,
                    self.settings,
                )
                self.g_infeasibility.add(
                    sol["x"][self.idx_x_integer], 0, multiplier * dx, grad_corr
                )
            else:
                self.g_infeasibility.add(
                    sol["x"][self.idx_x_integer], 0, multiplier * dx
                )
        else:
            self.g_infeasibility.add(sol["x"][self.idx_x_integer], 0, multiplier * dx)
        colored("Adding infeasible cut for closest point.", "blue")

    def _add_infeasible_cut(self, x_sol, lam_g_sol, nlpdata: MinlpData):
        """Add infeasibility cut using the default benders-inf cut method."""
        x_sol = x_sol[: self.nr_x_orig]
        g_k = self.g(x_sol, nlpdata.p)
        jac_g_k = self.jac_g_bin(x_sol, nlpdata.p)

        # g_k > ubg -> lam_g_sol > 0
        # Convert to g_lin < 0 -> g_k - ubg > 0
        g_bar_k = np.abs(lam_g_sol).T @ (
            (lam_g_sol > 0)
            * (g_k > nlpdata.ubg)
            * (g_k - np.where(np.isinf(nlpdata.ubg), 0, nlpdata.ubg))
            + (lam_g_sol < 0)
            * (g_k < nlpdata.lbg)
            * (g_k - np.where(np.isinf(nlpdata.lbg), 0, nlpdata.lbg))
        )
        # assert g_bar_k > 0
        if g_bar_k < self.settings.EPS:
            colored(f"Infeasibility cut of {g_bar_k}.")
        else:
            colored(f"Infeasibility cut of {g_bar_k}.", "blue")
        g_bar_k = max(g_bar_k, self.settings.EPS)
        # g_bar_k is positive by definition
        # + 10 is an extra tolerance.
        grad_g_bar_k = self.clip_gradient(g_bar_k + 10, 0, (lam_g_sol.T @ jac_g_k).T)

        x_sol_best_bin = self.sol_best["x"][self.idx_x_integer]
        x_bin_new = x_sol[self.idx_x_integer]
        if not self._check_cut_valid(
            g_bar_k, grad_g_bar_k, x_sol_best_bin, x_bin_new, 0.0
        ):
            # need gradient correction because we cut out best point
            grad_corr = compute_gradient_correction(
                x_sol_best_bin, x_bin_new, 0, g_bar_k, grad_g_bar_k, self.settings
            )
            self.g_infeasibility.add(x_bin_new, g_bar_k, grad_g_bar_k, grad_corr)
        else:
            self.g_infeasibility.add(x_bin_new, g_bar_k, grad_g_bar_k)
        colored("Adding standard infeasible cut.", "blue")

    def _add_oa(self, x_sol, nlpdata: MinlpData):
        x_sol = x_sol[: self.nr_x_orig]
        g_k = self.g(x_sol, nlpdata.p)
        jac_g_k = self.jac_g(x_sol, nlpdata.p)
        if self.idx_g_conv is not None:
            for i in self.idx_g_conv:
                if not np.isinf(nlpdata.ubg[i]):
                    self.g_oa_cvx_constraints.add(
                        x_sol, g_k[i] - nlpdata.ubg[i], jac_g_k[i, :].T
                    )
                else:
                    self.g_oa_cvx_constraints.add(
                        x_sol, -g_k[i] + nlpdata.lbg[i], -jac_g_k[i, :].T
                    )
            colored(f"Add {len(self.idx_g_conv)} OA cuts.", "blue")

    def _lowerapprox_oa(self, x, nlpdata):
        """Add outer approximation-based cuts for the objective value, the gradient is corrected when needed to keep the current best point feasible."""
        f_k = self.f(x, nlpdata.p)
        f_grad = self.grad_f_x(x, nlpdata.p)

        if not self._check_cut_valid(f_k, f_grad, self.sol_best["x"], x, self.y_N_val):
            grad_corr = compute_gradient_correction(
                self.sol_best["x"], x, self.y_N_val, f_k, f_grad, self.settings
            )
            self.g_oa_objective.add(x, f_k, f_grad, grad_corr)
        else:
            self.g_oa_objective.add(x, f_k, f_grad)

    def _gradient_correction(self, x_sol, lam_x_sol, nlpdata: MinlpData):
        """Correct gradient of the existing lower approximation cuts."""
        # TODO: (to improve computation speed) if the best point does not change, check only the last point
        x_sol_best_bin = self.sol_best["x"][self.idx_x_integer]

        # On the last integer point: check, correct (if needed) and add to g_lowerapprox
        f_k = self.f(x_sol, nlpdata.p)
        x_bin_new = x_sol[self.idx_x_integer]
        lambda_k = self.clip_gradient(
            f_k, self.internal_lb, -lam_x_sol[self.idx_x_integer]
        )

        if not self._check_cut_valid(
            f_k, lambda_k, x_sol_best_bin, x_bin_new, self.y_N_val
        ):
            grad_corr = compute_gradient_correction(
                x_sol_best_bin, x_bin_new, self.y_N_val, f_k, lambda_k, self.settings
            )
            self.g_benders.add(x_bin_new, f_k, lambda_k, grad_corr)
            logger.debug("Correcting new gradient at current best point")
        else:
            self.g_benders.add(x_bin_new, f_k, lambda_k)

    def _get_g_linearized(self, x, dx, nlpdata):
        if not self.sol_best_feasible and self.trust_region_fails:
            return Constraints()
        elif (self.g.size1_out("o0") == 0) & (self.g.size2_out("o0") == 0):
            if self.settings.WITH_DEBUG:
                breakpoint()
            return Constraints()
        else:
            g_lin = self.g(x, nlpdata.p)
            jac_g = self.jac_g(x, nlpdata.p)
            return Constraints(
                g_lin.numel(),
                (g_lin + jac_g @ dx),
                nlpdata.lbg - self.settings.EPS,
                nlpdata.ubg + self.settings.EPS,
            )

    def _solve_br_miqp_problem(self, nlpdata: MinlpData, constraint) -> MinlpData:
        """Solve BR-MIQP problem."""
        dx = self._x - self.sol_best["x"]

        if self.f_qp is None:
            f_k = self.f(self.sol_best["x"], nlpdata.p)
            f_lin = self.grad_f_x(self.sol_best["x"], nlpdata.p)
            f_hess = self.f_hess(self.sol_best["x"], nlpdata.p)
            if self.hessian_not_psd:
                min_eigen_value = np.linalg.eigh(f_hess.full())[0][0]
                if min_eigen_value < 0:
                    f_hess -= min_eigen_value * ca.DM.eye(self.nr_x_orig)
                    logger.info(
                        colored(
                            f"Negative eigenvalue detected {min_eigen_value}.", "red"
                        )
                    )
                else:
                    self.hessian_not_psd = False

            f = f_k + f_lin.T @ dx + 0.5 * dx.T @ f_hess @ dx
        else:
            f = self.f_qp(self._x, self.sol_best["x"], nlpdata.p)
        # Order seems to be important!
        g_cur_lin = self._get_g_linearized(self.sol_best["x"], dx, nlpdata)

        g_total = (
            g_cur_lin
            + self.g_benders
            + self.g_infeasibility
            + self.g_oa_objective
            + self.g_oa_cvx_constraints
        )

        available_time = max(
            1e-1,
            min(
                self.options_master[self.mip_timelimit_options_str],
                self.settings.TIME_LIMIT - toc(),
            ),
        )
        self.options[self.mip_timelimit_options_str] = available_time
        solver = ca.qpsol(
            f"benders_constraint_{self.g_benders.nr}",
            self.settings.MIP_SOLVER,
            {"f": f, "g": g_total.eq, "x": self._x, "p": self._nu},
            self.options,  # + {"error_on_fail": False}
        )

        solution = solver(
            x0=self.sol_best["x"],
            lbx=nlpdata.lbx,
            ubx=nlpdata.ubx,
            lbg=g_total.lb,
            ubg=g_total.ub,
            p=[constraint],
        )
        success, stats = self.collect_stats("BR-MIQP", solver, solution)
        if stats["return_status"] == "TIME_LIMIT" and not np.any(
            np.isnan(solution["x"].full())
        ):
            success = True
        return solution, success, stats

    def _solve_lb_milp_problem(self, nlpdata: MinlpData) -> MinlpData:
        """Solve LB-MILP problem - Benders master problem with one OA constraint."""
        dx = self._x - self.sol_best["x"]

        f_k = self.f(self.sol_best["x"], nlpdata.p)
        f_lin = self.grad_f_x(self.sol_best["x"], nlpdata.p)
        f = f_k + f_lin.T @ dx

        # Adding the following linearization might not be the best idea since
        # They can lead to false results!
        if not self.sol_best_feasible:
            g_cur_lin = Constraints()
        else:
            g_cur_lin = self._get_g_linearized(self.sol_best["x"], dx, nlpdata)
        g_total = (
            g_cur_lin
            + self.g_benders
            + self.g_infeasibility
            + self.g_oa_objective
            + self.g_oa_cvx_constraints
        )

        # Add extra constraint (one step OA):
        g_total.add(-ca.inf, f - self._nu, 0)
        g, ubg, lbg = g_total.eq, g_total.ub, g_total.lb

        available_time = max(
            1e-1,
            min(
                self.options_master[self.mip_timelimit_options_str],
                self.settings.TIME_LIMIT - toc(),
            ),
        )
        self.options_master[self.mip_timelimit_options_str] = available_time
        solver = ca.qpsol(
            f"benders_with_{self.g_benders.nr}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": self._nu,
                "g": g,
                "x": ca.vertcat(self._x, self._nu),
            },
            self.options_master,
        )

        solution = solver(
            x0=ca.vertcat(self.sol_best["x"], self.y_N_val + 1e-5),
            lbx=ca.vertcat(nlpdata.lbx, -ca.inf),
            ubx=ca.vertcat(nlpdata.ubx, self.y_N_val),
            lbg=lbg,
            ubg=ubg,
        )
        success, stats = self.collect_stats("LB-MILP", solver, solution)
        if stats["return_status"] == "TIME_LIMIT" and not np.any(
            np.isnan(solution["x"].full())
        ):
            colored(
                f"LB-MILP finished due to time limit, returning best incumbent solution (mip gap > {self.mipgap_milp})"
            )
            success = True
            solution["x"] = solution["x"][:-1]
        elif not success:
            if not self.sol_best_feasible:
                raise Exception("Problem can not be solved - Feasible zone is empty")
            solution = self.sol_best
            success = True
            colored("Failed solving LB-MILP.")
        else:
            solution["x"] = solution["x"][:-1]
            logger.info("SOLVED LB-MILP")

        return solution, success, stats

    def update_options(self, relaxed=False):
        """Update options."""
        if self.internal_lb > self.y_N_val:
            self.internal_lb = self.y_N_val

        if relaxed:
            if self.settings.USE_TIGHT_MIPGAP_FIRST_ITERATION:
                self.options[self.mipgap_options_str] = self.mipgap_miqp
            else:
                self.options[self.mipgap_options_str] = 1.0
        else:
            self.options[self.mipgap_options_str] = self.mipgap_miqp

        logger.info(
            f"MIP Gap set to {self.mipgap_miqp} - "
            f"Expected Range lb={self.internal_lb}  ub={self.y_N_val}"
        )

    def update_relaxed_solution(self, nlpdata: MinlpData):
        """Update the benders region using the relaxed solution."""
        for solved, sol in zip(nlpdata.solved_all, nlpdata.prev_solutions):
            if solved:
                # check if new best solution found
                if np.isinf(self.internal_lb) or self.internal_lb > float(sol["f"]):
                    if self.settings.USE_RELAXED_SOL_AS_LINEARIZATION:
                        # warm start with relaxed solution
                        self.sol_best["x"] = sol["x"][: self.nr_x_orig]
                    self.internal_lb = float(sol["f"])

        self._gradient_corrections_old_cuts()
        self.add_python_solver_time(toc())

    def _solve_mix(self, nlpdata: MinlpData):
        """Preparation for solving both Benders region master problem (BR-MIQP) and lower bound master problem (LB-MILP)."""
        # We miss the LB, try to find one...

        need_lb_milp = np.isinf(self.internal_lb) or (
            max(self.stats["BR-MIQP.iter"] - self.stats["best_iter"], 0)
            - max(self.stats["LB-MILP.iter"] - self.stats["best_iter"], 0)
            >= 3
        )

        if not need_lb_milp:
            constraint = self.internal_lb + self.alpha_kronqvist * (
                self.y_N_val - self.internal_lb
            )  # Kronqvist's trick
            solution, success, stats = self._solve_br_miqp_problem(nlpdata, constraint)

            if success:
                if solution["f"] > self.y_N_val:
                    nlpdata = get_solutions_pool(
                        nlpdata,
                        success,
                        stats,
                        self.settings,
                        solution,
                        self.idx_x_integer,
                    )
                    need_lb_milp = True
                if any_equal(
                    solution["x"],
                    [sol["x"] for sol in nlpdata.best_solutions],
                    self.idx_x_integer,
                ):
                    colored("BR-MIQP stagnates, need LB-MILP problem.", "yellow")
                    need_lb_milp = True
            else:
                self.trust_region_fails = True
                colored("Failed solving BR-MIQP.", "red")
                need_lb_milp = True

        if need_lb_milp:
            solution, success, stats = self._solve_lb_milp_problem(nlpdata)
            self.internal_lb = float(solution["f"])

        nlpdata = get_solutions_pool(
            nlpdata, success, stats, self.settings, solution, self.idx_x_integer
        )

        if not need_lb_milp:
            # Update obj val of br-miqp to internal_lb to avoid update in GenenericDecomposition
            for i in range(len(nlpdata.prev_solutions)):
                nlpdata.prev_solutions[i]["f"] = self.internal_lb
        return nlpdata

    def _solve_milp_only(self, nlpdata: MinlpData, integers_relaxed: bool):
        """Solve lower bound MILP only (LB-MILP)."""
        if integers_relaxed:
            solution, success, stats = self._solve_milp_from_relaxed_solution(nlpdata)
        else:
            solution, success, stats = self._solve_lb_milp_problem(nlpdata)
            self.internal_lb = float(solution["f"])
        return get_solutions_pool(
            nlpdata, success, stats, self.settings, solution, self.idx_x_integer
        )

    def _solve_milp_from_relaxed_solution(self, nlpdata: MinlpData):
        constraint = self.internal_lb + self.alpha_kronqvist * (
            self.y_N_val - self.internal_lb
        )  # Kronqvist's trick
        dx = self._x - self.sol_best["x"]
        f_k = self.f(self.sol_best["x"], nlpdata.p)
        f_lin = self.grad_f_x(self.sol_best["x"], nlpdata.p)
        f = f_k + f_lin.T @ dx
        # Order seems to be important!
        g_cur_lin = self._get_g_linearized(self.sol_best["x"], dx, nlpdata)

        g_total = (
            g_cur_lin
            + self.g_benders
            + self.g_infeasibility
            + self.g_oa_objective
            + self.g_oa_cvx_constraints
        )

        available_time = max(1e-1, self.settings.TIME_LIMIT - toc())
        self.options[self.mip_timelimit_options_str] = available_time
        self.options
        solver = ca.qpsol(
            "milp_from_relaxed_solution",
            self.settings.MIP_SOLVER,
            {"f": f, "g": g_total.eq, "x": self._x, "p": self._nu},
            self.options,  # + {"error_on_fail": False}
        )

        solution = solver(
            x0=self.sol_best["x"],
            lbx=nlpdata.lbx,
            ubx=nlpdata.ubx,
            lbg=g_total.lb,
            ubg=g_total.ub,
            p=[constraint],
        )
        success, stats = self.collect_stats("R-MILP", solver, solution)
        if stats["return_status"] == "TIME_LIMIT" and not np.any(
            np.isnan(solution["x"].full())
        ):
            success = True
        return solution, success, stats

    def _solve_tr_only(self, nlpdata: MinlpData):
        """Solve Benders region problem only (BR-MIQP)."""
        if self.internal_lb > -ca.inf:
            constraint = self.internal_lb + self.alpha_kronqvist * (
                self.y_N_val - self.internal_lb
            )  # Kronqvist's trick
        else:
            constraint = self.y_N_val * 0.99

        self.options[self.mipgap_options_str] = self.mipgap_miqp
        solution, success, stats = self._solve_br_miqp_problem(nlpdata, constraint)
        if success:
            nlpdata = get_solutions_pool(
                nlpdata, success, stats, self.settings, solution, self.idx_x_integer
            )
        else:
            if self.sol_best_feasible:
                solution = self.sol_best
                nlpdata.prev_solutions = [self.sol_best]
                nlpdata.solved_all = [True]
            else:
                Exception("Problem can not be solved")
        return nlpdata

    def compute_lb(self, x_sol):
        """Compute LB."""
        return np.max(self.g_benders(x_sol[self.idx_x_integer], 0))

    def update_sol_infeas(self, sol, obj_inf_sol):
        """Update if infeasible."""
        self.sol_infeasibility = obj_inf_sol
        colored(f"Infeasible solution new upper bound: {self.sol_infeasibility}", "red")
        if not self.sol_best_feasible:
            self.sol_best = sol

    def update_sol(self, sol):
        """Update solution."""
        self.sol_best_feasible = True
        self.sol_infeasibility = 0.0
        self.sol_best = sol
        self.y_N_val = float(sol["f"])
        self.early_benders = False
        colored(f"New upper bound: {self.y_N_val}", "green")

    def add_solutions(self, nlpdata: MinlpData, integers_relaxed=False):
        """Add solutions."""
        if integers_relaxed:
            self.update_relaxed_solution(nlpdata)
        else:
            needs_trust_region_update = False
            for solved, sol in zip(nlpdata.solved_all, nlpdata.prev_solutions):
                # check if new best solution found
                nonzero = np.count_nonzero(
                    (sol["x"][: self.nr_x_orig] - self.sol_best["x"])[
                        self.idx_x_integer
                    ]
                )
                if solved:
                    self._gradient_correction(sol["x"], sol["lam_x"], nlpdata)
                    needs_trust_region_update = True
                    if float(sol["f"]) + self.settings.EPS < self.y_N_val:
                        sol["x"] = sol["x"][: self.nr_x_orig]
                        self.update_sol(sol)
                    colored(f"Regular Cut {float(sol['f']):.3f} - {nonzero}.", "blue")
                    if self.with_oa_conv_cuts:
                        self._lowerapprox_oa(sol["x"], nlpdata)
                else:
                    if not self.sol_best_feasible:
                        if "x_infeasible" in sol:
                            # Use distance
                            infeas = ca.norm_2(
                                sol["x_infeasible"][: self.nr_x_orig]
                                - sol["x"][: self.nr_x_orig]
                            )
                        else:
                            # Use regular benders
                            g = sol["g"]
                            infeas = compute_infeasibility(
                                g.full(), nlpdata.lbg, nlpdata.ubg
                            )

                        if infeas < self.sol_infeasibility:
                            self.update_sol_infeas(sol, infeas)

                    colored(f"Infeasibility Cut - distance {nonzero}.", "blue")
                    self._add_infeasibility_cut(sol, nlpdata)

            if (
                self.with_oa_conv_cuts
            ):  # Add OA constraint cuts only for first solution in the pool to avoid slow down
                self._add_oa(nlpdata.prev_solutions[0]["x"], nlpdata)

            if needs_trust_region_update:
                self._gradient_amplification()

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """Solve."""
        self.add_solutions(nlpdata, integers_relaxed)
        self.add_python_solver_time(toc())

        self.update_options(integers_relaxed)
        if self.with_lb_milp:
            if self.with_milp_only:
                return self._solve_milp_only(nlpdata, integers_relaxed)
            else:
                return self._solve_mix(nlpdata)
        else:
            return self._solve_tr_only(nlpdata)

    def reset(self, nlpdata: MinlpData):
        """Reset data."""
        self.g_benders = LowerApproximation(
            self._x_bin, self._nu
        )  # Benders' cut: J_i + lambda_y_i.T (y - y_i) <= 0
        self.g_oa_objective = LowerApproximation(
            self._x, self._nu
        )  # OA cut for objective: f_i + grad_f_i (x - x_i) <= 0
        self.g_infeasibility = LowerApproximation(
            self._x_bin, 0
        )  # Infeasibility cut (y_hat - y_bar).T (y - y_bar) + sigma <= 0
        self.g_oa_cvx_constraints = LowerApproximation(
            self._x, 0
        )  # If g_i is cvx: g_i - ubg_i + jac_g_i.T (x - x_sol) <= 0
        self.internal_lb = -ca.inf
        self.sol_best_feasible = False
        self.sol_infeasibility = ca.inf
        self.sol_best = nlpdata._sol  # take a point as initialization
        self.y_N_val = 1e15  # Should be inf but can not at the moment ca.inf
        self.trust_region_fails = False

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart with new data."""
        relaxed = self.stats.relaxed_solution
        if relaxed:
            self.add_solutions(relaxed, True)

        if not nlpdata.relaxed:
            self.internal_lb = self.stats["lb"]
            self.add_solutions(nlpdata)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Master problem of the sequential Voronoi-based MIQP algorithm."""

import logging
import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    get_idx_linear_bounds,
    regularize_options,
    get_idx_inverse,
    extract_bounds,
)
from camino.settings import GlobalSettings, Settings
from camino.utils import colored
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class VoronoiTrustRegionMIQP(SolverClass):
    r"""
    Voronoi trust region problem.

    This implementation assumes the following input:
        min f(x)
        s.t. lb < g(x)

    It constructs the following problem:
        min f(x) + \nabla f(x) (x-x^i) + 0.5  (x-x^i)' (\nabla^2 f(x)) (x-x^i)
        s.t.
            lb \leq g(x) + \nabla  g(x) (x-x^i)
            NLPF constraints
            Voronoi trust region
    """

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create sequential Voronoi master problem."""
        super(VoronoiTrustRegionMIQP, self).__init__(problem, stats, s)
        self.options = regularize_options(s.MIP_SETTINGS, {}, s)

        self.f = ca.Function(
            "f", [problem.x, problem.p], [problem.f], {"jit": s.WITH_JIT}
        )
        self.grad_f = ca.Function(
            "gradient_f_x",
            [problem.x, problem.p],
            [ca.gradient(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )
        if problem.gn_hessian is not None:
            self.f_hess = ca.Function(
                "gn_hess_f_x", [problem.x, problem.p], [problem.gn_hessian]
            )
        else:
            self.f_hess = ca.Function(
                "hess_f_x",
                [problem.x, problem.p],
                [ca.hessian(problem.f, problem.x)[0]],
            )

        self.g = ca.Function(
            "g", [problem.x, problem.p], [problem.g], {"jit": s.WITH_JIT}
        )
        self.jac_g_bin = ca.Function(
            "jac_g_bin",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)[:, problem.idx_x_integer]],
            {"jit": s.WITH_JIT},
        )
        self.idx_g_lin = get_idx_linear_bounds(problem)
        self.idx_g_nonlin = get_idx_inverse(self.idx_g_lin, problem.g.shape[0])
        self.g_lin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_lin]],
            {"jit": s.WITH_JIT},
        )
        self.g_nonlin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_nonlin]],
            {"jit": s.WITH_JIT},
        )
        self.jac_g_nonlin = ca.Function(
            "gradient_g_x",
            [problem.x, problem.p],
            [ca.jacobian(problem.g[self.idx_g_nonlin], problem.x)],
            {"jit": s.WITH_JIT},
        )

        self._x = GlobalSettings.CASADI_VAR.sym("x_voronoi", problem.x.numel())
        self.idx_x_integer = problem.idx_x_integer
        self.nr_g_orig = problem.g.shape[0]
        self.nr_x = problem.x.shape[0]
        # Copy the linear constraints in g
        self.nr_g, self._g, self._lbg, self._ubg = extract_bounds(
            problem, data, self.idx_g_lin, self._x, allow_fail=False
        )

        self.options["discrete"] = [
            1 if i in self.idx_x_integer else 0 for i in range(self.nr_x)
        ]

        # Initialization for algorithm iterates
        self.ub = 1e15  # UB
        self.x_sol_list = []  # list of all the solution visited
        # Solution with best objective so far, we can safely assume the first
        # solution is the best even if infeasible since it is the only one
        # available yet.
        self.idx_best_x_sol = 0
        self.feasible_x_sol_list = []

    def add_solution(self, nlpdata, solved, solution, integers_relaxed=False):
        """Add a cut."""
        x_sol = solution["x"]
        obj_val = float(solution["f"])
        if x_sol.shape[0] == 1:
            x_sol = to_0d(x_sol)[np.newaxis]
        else:
            x_sol = to_0d(x_sol)[: self.nr_x]

        self.x_sol_list.append(x_sol)
        self.feasible_x_sol_list.append(solved)
        if solved and (not integers_relaxed):
            if obj_val < self.ub:
                self.ub = obj_val
                # TODO: a surrogate for counting iterates, it's a bit clutter
                self.idx_best_x_sol = len(self.x_sol_list) - 1
                logger.info(colored(f"New upperbound: {self.ub}", color="green"))
        else:
            g_k, lbg_k, ubg_k = self._generate_infeasible_cut(
                self._x, x_sol, solution["lam_g"], nlpdata.p
            )
            self._g = ca.vertcat(self._g, g_k)
            self._lbg = ca.vertcat(self._lbg, lbg_k)
            self._ubg = ca.vertcat(self._ubg, ubg_k)

    def solve(
        self, nlpdata: MinlpData, prev_feasible=True, is_qp=True, integers_relaxed=False
    ) -> MinlpData:
        """Solve sequential Voronoi master problem (MIQP)."""
        # Update with the lowest upperbound and the corresponding best solution:
        for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
            self.add_solution(nlpdata, solved, solution)

        x_sol_best = self.x_sol_list[self.idx_best_x_sol]

        # Create a new voronoi cut
        g_voronoi, lbg_voronoi, ubg_voronoi = self._generate_voronoi_tr(
            self._x[self.idx_x_integer], nlpdata.p
        )

        dx = self._x - x_sol_best

        f_k = self.f(x_sol_best, nlpdata.p)
        f_lin = self.grad_f(x_sol_best, nlpdata.p)
        if is_qp:
            f_hess = self.f_hess(x_sol_best, nlpdata.p)
            f = f_k + f_lin.T @ dx + 0.5 * dx.T @ f_hess @ dx
        else:
            f = f_k + f_lin.T @ dx

        g_lin = self.g_nonlin(x_sol_best, nlpdata.p)
        if g_lin.numel() > 0:
            jac_g = self.jac_g_nonlin(x_sol_best, nlpdata.p)
        else:
            g_lin = to_0d(g_lin)
            jac_g = np.zeros((0, self.nr_x))

        g = ca.vertcat(
            self._g,
            g_voronoi,
            g_lin + jac_g @ dx,
        )
        lbg = ca.vertcat(
            self._lbg,
            lbg_voronoi,
            nlpdata.lbg[self.idx_g_nonlin],
        )
        ubg = ca.vertcat(
            self._ubg,
            ubg_voronoi,
            nlpdata.ubg[self.idx_g_nonlin],
        )
        self.nr_g = ubg.numel()

        solver = ca.qpsol(
            f"voronoi_tr_milp_with_{self.nr_g}_constraints",
            self.settings.MIP_SOLVER,
            {
                "f": f,
                "g": g,
                "x": self._x,
            },
            self.options,
        )

        nlpdata.prev_solution = solver(
            x0=x_sol_best, lbx=nlpdata.lbx, ubx=nlpdata.ubx, lbg=lbg, ubg=ubg
        )

        nlpdata.solved, stats = self.collect_stats("VTR-MIQP", solver, solution)
        return nlpdata

    def reset(
        self, nlpdata: MinlpData
    ):  # TODO: to update, just copy paste from outer approx
        """Reset."""
        if self.idx_g_lin.numel() > 0:
            self.nr_g, self._g, self._lbg, self._ubg = extract_bounds(
                self.problem,
                nlpdata,
                self.idx_g_lin,
                self._x,
                self.problem.idx_x_integer,
            )
        else:
            self.nr_g, self._g, self._lbg, self._ubg = 0, [], [], []

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart algorithm."""
        relaxed = self.stats.relaxed_solution
        if relaxed:
            self.add_solution(nlpdata, True, relaxed.solutions_all[0], True)
        if not nlpdata.relaxed:
            for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
                self.add_solution(nlpdata, solved, solution)

    def _generate_infeasible_cut(self, x, x_sol, lam_g, p):
        """Generate infeasibility cut."""
        h_k = self.g(x_sol, p)
        jac_h_k = self.jac_g_bin(x_sol, p)
        g_k = lam_g.T @ (
            h_k + jac_h_k @ (x[self.idx_x_integer] - x_sol[self.idx_x_integer])
        )
        return g_k, -ca.inf, 0.0

    def _generate_voronoi_tr(self, x_bin, p):
        r"""
        Generate Voronoi trust region based on the visited integer solutions and the best integer solution so far.

        :param p: parameters
        """
        g_k = []
        lbg_k = []
        ubg_k = []

        x_sol_bin_best = self.x_sol_list[self.idx_best_x_sol][self.idx_x_integer]
        x_sol_bin_best_norm2_squared = x_sol_bin_best.T @ x_sol_bin_best
        for x_sol, is_feas in zip(self.x_sol_list, self.feasible_x_sol_list):
            x_sol_bin = x_sol[self.idx_x_integer]
            if is_feas and not np.allclose(x_sol_bin, x_sol_bin_best):
                a = ca.DM(2 * (x_sol_bin - x_sol_bin_best))
                b = ca.DM(x_sol_bin.T @ x_sol_bin - x_sol_bin_best_norm2_squared)
                g_k.append(a.T @ x_bin - b)
                lbg_k.append(-np.inf)
                ubg_k.append(0)

        return ca.vcat(g_k), lbg_k, ubg_k



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""A generic simple solver."""

from abc import ABC, abstractmethod
from copy import copy
from typing import List, Optional
from camino.problems import MinlpProblem, MinlpData
from camino.settings import Settings
from camino.utils import colored
from camino.stats import Stats
import casadi as ca
import numpy as np
import logging

from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class SolverClass(ABC):
    """Create solver class."""

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create a solver class."""
        self.settings = s
        self.stats = stats
        self.solver = None

    @abstractmethod
    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve the problem."""

    def collect_stats(self, algo_name, solver=None, sol=None):
        """Collect statistics."""
        logger.info(f"Solved {algo_name}")
        if solver is None:
            stats = self.solver.stats()
        else:
            stats = solver.stats()

        t_proc = sum([v for k, v in stats.items() if "t_proc" in k])

        t_wall = sum([v for k, v in stats.items() if "t_wall" in k])
        self.stats[f"{algo_name}.time"] += t_proc
        self.stats[f"{algo_name}.time_wall"] += t_wall
        self.stats[f"{algo_name}.iter"] += max(
            stats.get("n_call_solver", 0), stats["iter_count"]
        )
        self.stats[f"{algo_name}.runs"] += 1
        self.stats["t_solver_total"] += max(t_wall, t_proc)
        self.stats["success"] = stats["success"]
        self.stats["iter_type"] = algo_name
        if sol is not None:
            self.stats["sol_x"] = to_0d(sol["x"])
            self.stats["sol_obj"] = float(sol["f"])
        if self.settings.WITH_LOG_DATA:
            self.stats.save()
        return stats["success"], stats

    def add_python_solver_time(self, time):
        """Add solver time."""
        self.stats["t_python_solver"] += time


class MiSolverClass(SolverClass):

    def __init__(self, problem, data, stats, settings):
        """Create a generic MiSolverClass."""
        SolverClass.__init__(self, problem, stats, settings)
        self.stats["lb"] = -ca.inf
        self.stats["ub"] = ca.inf
        self.stats["iter_nr"] = 0
        self.stats["best_iter"] = -1
        self.best_solutions = []

    def update_best_solutions(self, data: MinlpData):
        """Update best solutions,"""
        if np.any(data.solved_all):
            for i, success in enumerate(data.solved_all):
                obj_val = float(data.prev_solutions[i]["f"])
                if success:
                    if obj_val + self.settings.EPS < self.stats["ub"]:
                        logger.info(
                            f"Decreased UB from {self.stats['ub']} to {obj_val}"
                        )
                        self.stats["ub"] = obj_val
                        self.best_solutions.append(data.prev_solutions[i])
                        self.stats["best_iter"] = copy(self.stats["iter_nr"])
                    elif obj_val - self.settings.EPS < self.stats["ub"]:
                        self.best_solutions.append(data.prev_solutions[i])

        if len(self.best_solutions) > 0:
            return self.best_solutions[-1]
        else:
            return None

    def get_best_solutions(self, data: MinlpData):
        """Get best solutions."""
        data.best_solutions = self.best_solutions
        if len(data.best_solutions) > 0:
            data.prev_solutions = [data.best_solutions[-1]]
            data.solved_all = [True]
        else:
            data.solved_all = [False] * len(data.solutions_all)
        return data

    @abstractmethod
    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve the problem."""

    @abstractmethod
    def reset(self, nlpdata: MinlpData):
        """Reset problem data."""

    @abstractmethod
    def warmstart(self, nlpdata: MinlpData):
        """Warmstart the algorithm."""


def regularize_options(options, default, s: Settings):
    """Regularize options."""
    ret = {} if options is None else options.copy()

    if not s.WITH_DEBUG:
        ret.update({"verbose": False, "print_time": 0})

    ret.update(default)

    return ret


def inspect_problem(problem: MinlpProblem, data: MinlpData):
    """Inspect problem for linear and convex bounds."""
    x_cont_idx = [
        i for i in range(problem.x.shape[0]) if i not in problem.idx_x_integer
    ]
    g_expr = ca.Function("g_func", [problem.x, problem.p], [problem.g])
    sp = np.array(g_expr.sparsity_jac(0, 0))

    g_lin = []
    g_lin_bin = []
    g_other = []
    g_conv = []
    for i in range(problem.g.shape[0]):
        hess = ca.hessian(problem.g[i], problem.x)[0]
        if hess.nnz() == 0:
            g_lin.append(i)
            if sum(sp[i, x_cont_idx]) == 0:
                g_lin_bin.append(i)
        elif not np.isinf(data.ubg[i]) and not np.isinf(data.lbg[i]):
            g_other.append(i)
        else:
            # NOTE: strong assumption! Works only for problems coming from the MINLPlib
            # A possible implementation is:
            # ccs = hess.sparsity().get_ccs()
            # hess[0, ccs[0]].is_numeric()
            g_conv.append(i)

    return g_lin, g_lin_bin, g_other, g_conv


def set_constraint_types(problem, g_lin, g_lin_bin, g_other, g_conv):
    """Set problem indices."""
    problem.idx_g_lin = g_lin
    problem.idx_g_lin_bin = g_lin_bin
    problem.idx_g_other = g_other
    problem.idx_g_conv = g_conv


def get_idx_linear_bounds_binary_x(problem: MinlpProblem):
    """Get the indices for the linear bounds that are purely on the binary x."""
    if problem.idx_g_lin_bin is None:
        x_cont_idx = [
            i for i in range(problem.x.shape[0]) if i not in problem.idx_x_integer
        ]
        g_expr = ca.Function("g_func", [problem.x, problem.p], [problem.g])
        sp = np.array(g_expr.sparsity_jac(0, 0))

        iterator = get_idx_linear_bounds(problem)
        problem.idx_g_lin_bin = np.array(
            list(
                filter(
                    lambda i: (
                        sum(sp[i, problem.idx_x_integer]) > 0
                        and sum(sp[i, x_cont_idx]) == 0
                    ),
                    iterator,
                )
            )
        )

    return problem.idx_g_lin_bin


def get_lin_bounds(problem: MinlpProblem):
    get_idx_linear_bounds_binary_x(problem)
    return problem.idx_g_lin, problem.idx_g_lin_bin


def get_idx_linear_bounds(problem: MinlpProblem):
    """Get the indices of the linear bounds."""
    if problem.idx_g_lin is None:
        nr_g = problem.g.shape[0]
        problem.idx_g_lin = np.array(
            list(
                filter(
                    lambda i: ca.hessian(problem.g[i], problem.x)[0].nnz() == 0,
                    range(nr_g),
                )
            )
        )

    return problem.idx_g_lin


def get_idx_inverse(indices, nr):
    """Get an inver list of indices."""
    full_indices = np.arange(0, nr)
    return list(set(full_indices) - set(indices))


def extract_bounds(
    problem: MinlpProblem,
    data: MinlpData,
    idx_g: List[int],
    new_x,
    idx_x: Optional[List[int]] = None,
    allow_fail=True,
):
    """Extract bounds."""
    empty = False
    nr_g = len(idx_g)
    if nr_g == 0:
        empty = True
    else:
        try:
            if idx_x is None:
                _x = problem.x
                g = ca.Function("g_lin", [_x, problem.p], [problem.g[idx_g]])(
                    new_x, data.p
                )
            else:
                vec = []
                j = 0
                for i in range(problem.x.shape[0]):
                    if i in idx_x:
                        vec.append(new_x[j])
                        j += 1
                    else:
                        vec.append(0)
                vec = ca.vertcat(*vec)
                vec_fn = ca.Function("v", [new_x], [vec])
                g = ca.Function("g_lin", [problem.x, problem.p], [problem.g[idx_g]])(
                    vec_fn(new_x), data.p
                )

            lbg = data.lbg[idx_g].flatten().tolist()
            ubg = data.ubg[idx_g].flatten().tolist()
        except Exception as e:
            if allow_fail:
                logger.warning(str(e))
                empty = True
            else:
                raise e

    if empty:
        return 0, [], [], []
    else:
        return nr_g, g, lbg, ubg



#!/usr/bin/python3
# PYTHON_ARGCOMPLETE_OK

# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

from sys import argv
from shutil import copyfile
from argparse import ArgumentParser
from camino.runner import runner, batch_runner
from camino.utils import setup_logger, logging
from camino.settings import Settings
import argcomplete


def main(args):
    """Process the data."""
    if Settings.WITH_DEBUG:
        setup_logger(logging.DEBUG)
    else:
        setup_logger(logging.INFO)

    parser = ArgumentParser(description="Benders solver")
    subparser = parser.add_subparsers(
        title="Commands",
        dest="command",
        description="",
        help="""for argument information use:
                                      <command-name> -h""",
    )
    parser_run = subparser.add_parser("run", help="Run a pre-defined problem")
    parser_run.add_argument("solver")
    parser_run.add_argument("problem_name")
    parser_run.add_argument("--save", default=None, type=str)
    parser_run.add_argument("--args", default=None, nargs="*", type=str)
    parser_copy = subparser.add_parser("copy", help="Copy NL files")
    parser_copy.add_argument("target")
    parser_copy.add_argument("solution")
    parser_copy.add_argument("nlfile")
    parser_batch = subparser.add_parser("batch", help="Run a batch of NL files")
    parser_batch.add_argument("algorithm")
    parser_batch.add_argument("target")
    parser_batch.add_argument("nlfiles", type=str, nargs="+")

    argcomplete.autocomplete(parser)
    parsed = parser.parse_args(args)

    if parsed.command == "copy":
        copyfile(parsed.nlfile, parsed.target)
        print(f"File copied to {parsed.target}")
    elif parsed.command == "batch":
        batch_runner(parsed.algorithm, parsed.target, parsed.nlfiles)
    elif parsed.command == "run":
        runner(parsed.solver, parsed.problem_name, parsed.save, parsed.args)
    else:
        parser.print_help()

    exit(0)


if __name__ == "__main__":
    main(argv[1:])



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Set of utilities to model a problem."""

from typing import Union
from camino.settings import GlobalSettings as GS
import casadi as ca


def integrate_rk4(
    x: GS.CASADI_VAR,
    u: GS.CASADI_VAR,
    x_dot: GS.CASADI_VAR,
    dt: Union[float, GS.CASADI_VAR],
    m_steps: int = 1,
):
    """
    Implement RK4 integrator for ODE.

    x: state
    u: control
    x_dot: ODE that describes the continuous time dynamics of the system
    dt: integration time
    m_steps: number of integration steps per interval
    """

    f = ca.Function("f", [x, u], [x_dot])
    X0 = GS.CASADI_VAR.sym("X0", x.shape[0])
    U = GS.CASADI_VAR.sym("U", u.shape[0])
    X = X0
    if isinstance(dt, GS.CASADI_VAR):
        DT = GS.CASADI_VAR.sym("DT")
    else:
        DT = dt
    dt_scaled = DT / m_steps
    for _ in range(m_steps):
        k1 = f(X, U)
        k2 = f(X + dt_scaled / 2 * k1, U)
        k3 = f(X + dt_scaled / 2 * k2, U)
        k4 = f(X + dt_scaled * k3, U)
        X += dt_scaled / 6 * (k1 + 2 * k2 + 2 * k3 + k4)

    if isinstance(dt, GS.CASADI_VAR):
        return ca.Function("I_rk4", [X0, U, DT], [X], ["x0", "u", "dt"], ["xf"])
    else:
        return ca.Function("I_rk4", [X0, U], [X], ["x0", "u"], ["xf"])


def integrate_ee(
    x: GS.CASADI_VAR,
    u: GS.CASADI_VAR,
    x_dot: GS.CASADI_VAR,
    dt: Union[float, GS.CASADI_VAR],
    m_steps: int = 1,
):
    """
    Implement explicit euler integrator for ODE.

    x: state
    u: control
    x_dot: ODE that describes the continuous time dynamics of the system
    dt: integration time
    m_steps: number of integration steps per interval
    """

    f = ca.Function("f", [x, u], [x_dot])
    X0 = GS.CASADI_VAR.sym("X0", x.shape[0])
    U = GS.CASADI_VAR.sym("U", u.shape[0])
    X = X0
    if isinstance(dt, GS.CASADI_VAR):
        DT = GS.CASADI_VAR.sym("DT")
    else:
        DT = dt
    for _ in range(m_steps):
        k1 = f(X, U)
        X += DT * k1 / m_steps
    if isinstance(dt, GS.CASADI_VAR):
        return ca.Function("I_ee", [X0, U, DT], [X], ["x0", "u", "dt"], ["xf"])
    else:
        return ca.Function("I_ee", [X0, U], [X], ["x0", "u"], ["xf"])


def integrate_ie(
    x: GS.CASADI_VAR,
    u: GS.CASADI_VAR,
    x_dot: GS.CASADI_VAR,
    dt: Union[float, GS.CASADI_VAR],
    m_steps: int = 1,
):
    """Integrate Implicit Euler."""
    f = ca.Function("f", [x, u], [x_dot])
    X0 = GS.CASADI_VAR.sym("X0", x.shape[0])
    Xk = GS.CASADI_VAR.sym("Xk", x.shape[0])
    U = GS.CASADI_VAR.sym("U", u.shape[0])
    if isinstance(dt, GS.CASADI_VAR):
        DT = GS.CASADI_VAR.sym("DT")
    else:
        DT = dt
    X = X0
    for _ in range(m_steps):
        X += DT * f(X, U) / m_steps
    f = X - Xk
    if isinstance(dt, GS.CASADI_VAR):
        return ca.Function(
            "I_ie", [X0, U, Xk, DT], [f], ["x0", "u", "x1", "dt"], ["xf"]
        )
    else:
        return ca.Function("I_ie", [X0, U, Xk], [f], ["x0", "u", "x1"], ["xf"])



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Implementation of the CIA algorithm."""

import copy
import datetime
import casadi as ca
import numpy as np
from typing import Tuple
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.solvers import SolverClass, Stats, MinlpProblem, MinlpData
from camino.settings import Settings
from camino.utils import toc, logging
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


def to_list(dt, min_time, nr_b):
    """Create a min up or downtime list."""
    if isinstance(min_time, int):
        return [dt * min_time for _ in range(nr_b)]
    else:
        return [dt * min_time[i] for i in range(nr_b)]


class SurSolver(SolverClass):
    """
    Create a solver to approximate relaxed integer sequences
    with the sum-up-rounding scheme.

    ASSUMPTION: uniform discretization grid!
    """

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create NLP problem."""
        super(SurSolver, self).__init__(problem, stats, s)
        self.idx_x_integer = problem.idx_x_integer
        self.dt = self.meta.dt
        self.meta = copy.deepcopy(problem.meta)

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve NLP."""
        b_rel = to_0d(nlpdata.x_sol)[self.idx_x_integer]
        if any(b_rel) < 0 or any(b_rel) > 1:
            raise ValueError("Spotted a relaxed binary out of range.")
        b_rel[b_rel < 0] = 0
        b_rel[b_rel > 1.0] = 1

        b_bin = np.zeros_like(b_rel)
        for i in range(b_bin.shape[1]):  # for each control
            marker = 0
            for t in range(b_bin.shape[0]):  # scam the horizon
                if (
                    np.sum(b_rel[i][-(t - marker) :] * self.dt)
                    - np.sum(b_bin[i][-(t - marker) : -1] * self.dt)
                    >= 0.5 * self.dt
                ):
                    b_bin[i][t] = 1
                else:
                    b_bin[i][t] = 0

        nlpdata.x_sol[self.idx_x_integer] = b_bin

        return nlpdata


if __name__ == "__main__":
    test_array = np.array(
        [
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
            [0.1, 0.1, 0.1, 0.1, 0.1, 0.1],
            [0.1, 0.1, 0.1, 0.5, 0.1, 0.1],
        ]
    )
    test_array = test_array.T

    # Add check for tall matrix shape
    dt = 1
    b_rel = test_array
    b_rel = np.atleast_2d(b_rel)
    if any(b_rel.flatten()) < -1e-3 or any(b_rel.flatten()) > 1 + 1e-3:
        raise ValueError("Spotted a relaxed binary out of range.")
    b_rel[b_rel < 0] = 0
    b_rel[b_rel > 1.0] = 1

    b_bin = np.ones_like(b_rel) * 99
    for b_bin_i, b_rel_i in zip(b_bin.T, b_rel.T):  # for each control
        marker = -1
        for t in range(b_bin_i.shape[0]):  # scam the horizon
            breakpoint()
            if t == 0:
                res = b_rel_i[0]
            else:
                res = np.sum(b_rel_i[marker : t + 1] * dt) - np.sum(
                    b_bin_i[marker:t] * dt
                )
            print(f"Res. {res}")
            if res >= 0.5 * dt:
                b_bin_i[t] = 1
                marker = 0
            else:
                b_bin_i[t] = 0
                marker += 1
    print(f"{b_bin=}")
    breakpoint()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Class of decomposition solvers."""

import casadi as ca
import numpy as np
from camino.solvers import MiSolverClass, Stats, MinlpProblem, MinlpData, Settings
from camino.solvers.utils import get_termination_condition
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.utils import colored, logging, toc, tic
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class GenericDecomposition(MiSolverClass):

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        master: MiSolverClass,
        fnlp: NlpSolver,
        termination_type: str = "std",
        first_relaxed: bool = True,
    ):
        """Generic decomposition algorithm."""
        super(GenericDecomposition, self).__init__(problem, data, stats, settings)
        self.termination_condition = get_termination_condition(
            termination_type, problem, data, settings
        )
        self.master = master
        self.nlp = NlpSolver(problem, stats, settings)
        self.fnlp = fnlp
        self.settings = settings
        self.stats = stats
        self.first_relaxed = first_relaxed
        self._x_nan = np.nan * np.empty(data.x0.shape[0])

    def solve(self, data: MinlpData, *args, **kwargs) -> MinlpData:
        """Solve the problem."""
        logger.info("Solver initialized.")
        tic()
        # Benders algorithm
        feasible = True
        x_hat = np.nan * np.empty(data.x0.shape[0])

        if self.first_relaxed:
            integers_relaxed = True
        else:
            integers_relaxed = False

        while (
            not self.termination_condition(
                self.stats,
                self.settings,
                self.stats["lb"],
                self.stats["ub"],
                self._get_x_star(),
                x_hat,
            )
        ) and feasible:
            # Solve NLP(y^k)
            data = self.nlp.solve(data, integers_relaxed=integers_relaxed)
            if integers_relaxed:
                self.stats["lb"] = data.obj_val

            # Is there any infeasible?
            if not np.all(data.solved_all):
                # Solve NLPF(y^k)
                data = self.fnlp.solve(data)
                logger.info(colored("Feasibility NLP solved.", "yellow"))

            if not integers_relaxed:
                self.update_best_solutions(data)

            # Solve master^k and set lower bound:
            data = self.master.solve(data, integers_relaxed=integers_relaxed)
            feasible = data.solved
            self.stats["lb"] = max(data.obj_val, self.stats["lb"])
            x_hat = data.x_sol
            logger.debug(
                f"x_hat = {to_0d(x_hat).tolist() if len(to_0d(x_hat).tolist()) < 5 else  to_0d(x_hat).tolist()[:5]} ..."
            )
            logger.debug(f"{self.stats['ub']=}, {self.stats['lb']=}\n")

            if self.stats["iter_nr"] == 0 and integers_relaxed == True:
                integers_relaxed = False

            self.stats["iter_nr"] += 1

        self.stats["total_time_calc"] = toc(reset=True)
        return self.get_best_solutions(data)

    def _get_x_star(self):
        if len(self.best_solutions) == 0:
            return self._x_nan
        else:
            return self.best_solutions[-1]["x"]

    def reset(self, nlpdata: MinlpData):
        """Reset Solvers."""
        self.master.reset(nlpdata)
        nlpdata.best_solutions = []

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart procedure."""
        if not nlpdata.relaxed:
            self.first_relaxed = False
            self.update_best_solutions(nlpdata)
            self.master.warmstart(nlpdata)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Set of benders-based solvers."""

from camino.utils import toc
from camino.solvers import Stats, MinlpProblem, MinlpData, Settings
from camino.solvers.subsolvers.fnlp import FeasibilityNlpSolver
from camino.solvers.subsolvers.fnlp_closest import FindClosestNlpSolver
from camino.solvers.decomposition.base import GenericDecomposition
from camino.solvers.decomposition.benders_master import (
    BendersMasterMILP,
    BendersMasterMIQP,
)
from camino.solvers.decomposition.oa_master import (
    OuterApproxMILP,
    OuterApproxMIQP,
    OuterApproxMILPImproved,
    OuterApproxMIQPImproved,
)
from camino.solvers.decomposition.voronoi_master import VoronoiTrustRegionMIQP
from camino.solvers.decomposition.sequential_benders_master import BendersRegionMasters


class GeneralizedBenders(GenericDecomposition):
    """Generalized Benders."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = BendersMasterMILP(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class GeneralizedBendersQP(GenericDecomposition):
    """Generalized Benders with an additional hessian in the cost function."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings = None,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = BendersMasterMIQP(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=True,
        )
        stats["total_time_loading"] = toc(reset=True)


class OuterApproximation(GenericDecomposition):
    """Linear Outer Approximation with standard feasibility restoration"""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings = None,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = OuterApproxMILP(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class OuterApproximationQP(GenericDecomposition):
    """Quadratic Outer Approximation with standard feasibility restoration"""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings = None,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = OuterApproxMIQP(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class OuterApproximationImproved(GenericDecomposition):
    """Improve Outer Approximation by adding cuts only on linear constraint to avoid infeasibility"""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings = None,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = OuterApproxMILPImproved(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class OuterApproximationQPImproved(GenericDecomposition):
    """Improve quadratic Outer Approximation by adding cuts only on linear constraint to avoid infeasibility"""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings = None,
        termination_type="std",
    ):
        """Generic decomposition algorithm."""
        master = OuterApproxMIQPImproved(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class SequentialVoronoiMIQP(GenericDecomposition):
    """Generalized Benders."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        termination_type="equality",
    ):
        """Generic decomposition algorithm."""
        master = VoronoiTrustRegionMIQP(problem, data, stats, settings)
        fnlp = FeasibilityNlpSolver(problem, data, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=False,
        )
        stats["total_time_loading"] = toc(reset=True)


class SequentialBendersMIQP(GenericDecomposition):
    """Generalized Benders."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        termination_type="std",
        with_lb_milp=True,
        with_milp_only=False,
    ):
        """Generic decomposition algorithm."""
        master = BendersRegionMasters(
            problem,
            data,
            stats,
            settings,
            with_lb_milp=with_lb_milp,
            with_milp_only=with_milp_only,
        )
        fnlp = FindClosestNlpSolver(problem, stats, settings)
        GenericDecomposition.__init__(
            self,
            problem,
            data,
            stats,
            settings,
            master,
            fnlp,
            termination_type,
            first_relaxed=True,
        )
        stats["total_time_loading"] = toc(reset=True)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Set of master solvers based on bender cuts.

Bender cuts are based on the principle of decomposing the problem into two
parts where the main part is only solving the integer variables.
"""

import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    get_idx_linear_bounds,
    get_idx_linear_bounds_binary_x,
    regularize_options,
    get_idx_inverse,
    extract_bounds,
)
from camino.settings import GlobalSettings, Settings
import logging

logger = logging.getLogger(__name__)


class BendersMasterMILP(SolverClass):
    """Create benders master problem."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        with_lin_bounds=True,
    ):
        """Create benders master MILP."""
        super(BendersMasterMILP, self).__init__(problem, stats, s)
        self.setup_common(problem, s)
        self.jac_g_bin = ca.Function(
            "jac_g_bin",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)[:, problem.idx_x_integer]],
            {"jit": s.WITH_JIT},
        )
        self._x = GlobalSettings.CASADI_VAR.sym("x_bin", self.nr_x_bin)
        self.problem = problem

        if with_lin_bounds:
            self.idx_g_lin = get_idx_linear_bounds_binary_x(problem)
            self.nr_g, self._g, self._lbg, self._ubg = extract_bounds(
                problem, data, self.idx_g_lin, self._x, problem.idx_x_integer
            )
        else:
            self.idx_g_lin = np.array([])
            self.nr_g, self._g, self._lbg, self._ubg = 0, [], [], []

        self.cut_id = 0

    def reset(self, data):
        """Reset."""
        if self.idx_g_lin.numel() > 0:
            self.nr_g, self._g, self._lbg, self._ubg = extract_bounds(
                self.problem, data, self.idx_g_lin, self._x, self.problem.idx_x_integer
            )
        else:
            self.nr_g, self._g, self._lbg, self._ubg = 0, [], [], []

        self.cut_id = 0

    def warmstart(self, data):
        """Warmstart algorithm."""
        if not data.relaxed:
            for solution in data.best_solutions:
                self.add_solution(
                    data, solution["x"], solution["lam_g"], solution["lam_x"], True
                )

    def setup_common(self, problem: MinlpProblem, s: Settings):
        """Set up common data."""
        self.settings = s
        self.options = regularize_options(s.MIP_SETTINGS, {}, s)

        self.f = ca.Function(
            "f", [problem.x, problem.p], [problem.f], {"jit": s.WITH_JIT}
        )
        self.g = ca.Function(
            "g", [problem.x, problem.p], [problem.g], {"jit": s.WITH_JIT}
        )

        self.idx_x_integer = problem.idx_x_integer
        self.nr_x_bin = len(problem.idx_x_integer)
        self._nu = GlobalSettings.CASADI_VAR.sym("nu", 1)
        self.nr_g_orig = problem.g.shape[0]
        self.nr_x_orig = problem.x.shape[0]
        self.options["discrete"] = np.ones_like(problem.idx_x_integer).tolist() + [0]

    def _generate_cut_equation(
        self, x, x_sol, x_sol_sub_set, lam_g, lam_x, p, lbg, ubg, prev_feasible
    ):
        r"""
        Generate a cut.

        when feasible, this creates the cut:

        $$
         g_k = f_k \lambda_k^{\mathrm{T}} (x_{\mathrm{binary}} - x_{\mathrm{binary, prev\_sol}}) - \nu \leq 0 \\
        \lambda_k = \nabla f_k + J_{g_k}^{T} \lambda_{g_j}
        $$

        when infeasible, the cut is equal to
        $$
        g_k = \lambda_{g_k}^{\mathrm{T}} g_k + J_{g_k}(x_{\mathrm{binary}} - x_{\mathrm{binary, prev\_sol}})
        $$

        :param x: optimization variable
        :param x_sol: Complete x_solution
        :param x_sol_sub_set: Subset of the x solution to optimize the MILP to
            $x_{\mathrm{bin}}$
        :param lam_g: Lambda g solution
        :param p: parameters
        :param prev_feasible: If the previous solution was feasible
        :return: g_k the new cutting plane (should be > 0)
        """
        if prev_feasible:
            lambda_k = -lam_x[self.idx_x_integer]
            f_k = self.f(x_sol, p)
            g_k = f_k + lambda_k.T @ (x - x_sol_sub_set) - self._nu
        else:  # Not feasible solution
            h_k = self.g(x_sol, p)
            jac_h_k = self.jac_g_bin(x_sol, p)
            g_k = lam_g.T @ (
                h_k
                + jac_h_k @ (x - x_sol_sub_set)
                - (lam_g > 0) * np.where(np.isinf(ubg), 0, ubg)
                + (lam_g < 0) * np.where(np.isinf(lbg), 0, lbg)
            )

        return g_k

    def add_solution(self, nlpdata, x_sol, lam_g_sol, lam_x_sol, prev_feasible):
        """Create cut."""
        g_k = self._generate_cut_equation(
            self._x,
            x_sol[: self.nr_x_orig],
            x_sol[self.idx_x_integer],
            lam_g_sol,
            lam_x_sol,
            nlpdata.p,
            nlpdata.lbg,
            nlpdata.ubg,
            prev_feasible,
        )
        self.cut_id += 1

        self._g = ca.vertcat(self._g, g_k)
        self._ubg.append(0)
        self._lbg.append(-ca.inf)
        self.nr_g += 1

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """solve."""
        x_bin_star = nlpdata.x_sol[self.idx_x_integer]
        if not integers_relaxed:
            self.add_solution(
                nlpdata,
                nlpdata.x_sol,
                nlpdata.lam_g_sol,
                nlpdata.lam_x_sol,
                nlpdata.solved,
            )
        solver = ca.qpsol(
            f"benders_with_{self.nr_g}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": self._nu,
                "g": self._g,
                "x": ca.vertcat(self._x, self._nu),
            },
            self.options,
        )

        # This solver solves only to the binary variables (_x)!
        solution = solver(
            x0=ca.vertcat(x_bin_star, nlpdata.obj_val),
            # NOTE harmonize lb with OA, here -1e5, in OA -1e8
            lbx=ca.vertcat(nlpdata.lbx[self.idx_x_integer], -1e5),
            ubx=ca.vertcat(nlpdata.ubx[self.idx_x_integer], ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        x_full = nlpdata.x_sol.full()[: self.nr_x_orig]
        x_full[self.idx_x_integer] = solution["x"][:-1]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("BENDERS-MILP", solver, solution)
        return nlpdata


class BendersMasterMIQP(BendersMasterMILP):
    """MIQP implementation."""

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create the benders constraint MILP."""
        super(BendersMasterMIQP, self).__init__(problem, data, stats, s)
        self.f_hess_bin = ca.Function(
            "hess_f_x_bin",
            [problem.x, problem.p],
            [
                ca.hessian(problem.f, problem.x)[0][problem.idx_x_integer, :][
                    :, problem.idx_x_integer
                ]
            ],
            {"jit": self.settings.WITH_JIT},
        )

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """solve."""
        x_bin_star = nlpdata.x_sol[self.idx_x_integer]
        if not integers_relaxed:
            g_k = self._generate_cut_equation(
                self._x,
                nlpdata.x_sol[: self.nr_x_orig],
                x_bin_star,
                nlpdata.lam_g_sol,
                nlpdata.lam_x_sol,
                nlpdata.p,
                nlpdata.lbg,
                nlpdata.ubg,
                nlpdata.solved,
            )
            self.cut_id += 1
            self._g = ca.vertcat(self._g, g_k)
            self._ubg.append(0)
            self._lbg.append(-ca.inf)
            self.nr_g += 1

        f_hess = self.f_hess_bin(nlpdata.x_sol[: self.nr_x_orig], nlpdata.p)

        dx = self._x - x_bin_star
        solver = ca.qpsol(
            f"benders_qp{self.nr_g}",
            self.settings.MIP_SOLVER,
            {
                "f": self._nu + 0.5 * dx.T @ f_hess @ dx,
                "g": self._g,
                "x": ca.vertcat(self._x, self._nu),
            },
            self.options,
        )

        # This solver solves only to the binary variables (_x)!
        solution = solver(
            x0=ca.vertcat(x_bin_star, nlpdata.obj_val),
            lbx=ca.vertcat(nlpdata.lbx[self.idx_x_integer], -1e5),
            ubx=ca.vertcat(nlpdata.ubx[self.idx_x_integer], ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        obj = solution["x"][-1].full()
        if obj > solution["f"]:
            raise Exception("Possible thougth mistake!")
        solution["f"] = obj
        x_full = nlpdata.x_sol.full()[: self.nr_x_orig]
        x_full[self.idx_x_integer] = solution["x"][:-1]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("BENDERS-MIQP", solver, solution)
        return nlpdata


class BendersTrustRegionMIP(BendersMasterMILP):
    """
    Create benders constraint MIP.

    By an idea of Moritz D. and Andrea G.
    Given the ordered sequence of integer solutions:
        Y := {y1, y2, ..., yN}
    such that J(y1) >= J(y2) >= ... >= J(yN) we define the
    benders polyhedral B := {y in R^n_y:
        J(y_i) + Nabla J(y_i)^T (y - y_i) <= J(y_N),
        forall i = 1,...,N-1
    }

    This MILP solves:
        min F(y, z | y_bar, z_bar)
        s.t ub >= H_L(y,z| y_bar, z_bar) >= lb
        with y in B

    For this implementation, since the original formulation implements:
        J(y_i) + Nabla J(yi) T (y - yi) <= nu,
        meaning: nu == J(y_N)
    """

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create the benders constraint MILP."""
        super(BendersTrustRegionMIP, self).__init__(problem, data, stats, s)
        self.setup_common(problem, s)

        self.idx_g_lin = get_idx_linear_bounds(problem)
        self.idx_g_nonlin = get_idx_inverse(self.idx_g_lin, problem.g.shape[0])

        self.grad_f_x_sub = ca.Function(
            "gradient_f_x",
            [problem.x, problem.p],
            [ca.gradient(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.jac_g_sub = ca.Function(
            "jac_g",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.f_hess = ca.Function(
            "hess_f_x", [problem.x, problem.p], [ca.hessian(problem.f, problem.x)[0]]
        )

        self._x = GlobalSettings.CASADI_VAR.sym("x_benders", problem.x.numel())
        self.nr_g, self._g, self._lbg, self._ubg = extract_bounds(
            problem, data, self.idx_g_lin, self._x, allow_fail=False
        )

        self.options.update(
            {
                "discrete": [
                    1 if elm in problem.idx_x_integer else 0
                    for elm in range(self._x.shape[0])
                ],
                "error_on_fail": False,
            }
        )
        self.y_N_val = 1e15  # Should be inf but can not at the moment ca.inf
        # We take a point
        self.best_data = None
        self.x_sol_best = data.x0

    def solve(self, nlpdata: MinlpData, relaxed=False) -> MinlpData:
        """Solve."""
        # Update with the lowest upperbound and the corresponding best solution:
        if relaxed:
            self.x_sol_best = nlpdata.x_sol[: self.nr_x_orig]
        elif nlpdata.obj_val < self.y_N_val and nlpdata.solved:
            self.y_N_val = nlpdata.obj_val
            self.x_sol_best = nlpdata.x_sol[: self.nr_x_orig]
            self.best_data = nlpdata._sol
            print(f"NEW BOUND {self.y_N_val}")

        # Create a new cut
        x_sol_prev = nlpdata.x_sol[: self.nr_x_orig]
        g_k = self._generate_cut_equation(
            self._x[self.idx_x_integer],
            x_sol_prev,
            x_sol_prev[self.idx_x_integer],
            nlpdata.lam_g_sol,
            nlpdata.lam_x_sol,
            nlpdata.p,
            nlpdata.lbg,
            nlpdata.ubg,
            nlpdata.solved,
        )
        self.cut_id += 1

        self._g = ca.vertcat(self._g, g_k)
        self._lbg = ca.vertcat(self._lbg, -ca.inf)
        # Should be at least 1e-4 better and 1e-4 from the constraint bound!
        self._ubg = ca.vertcat(self._ubg, 0)  # -1e-4)
        self.nr_g += 1

        dx = self._x - self.x_sol_best

        g_lin = self.g(self.x_sol_best, nlpdata.p)[self.idx_g_nonlin]
        if g_lin.numel() > 0:
            jac_g = self.jac_g_sub(self.x_sol_best, nlpdata.p)[self.idx_g_nonlin, :]
            # Warning: Reversing these expressions give weird results!
            g = ca.vertcat(
                g_lin + jac_g @ dx,
                self._g,
            )
            lbg = ca.vertcat(
                nlpdata.lbg[self.idx_g_nonlin],
                self._lbg,
            )
            ubg = ca.vertcat(
                nlpdata.ubg[self.idx_g_nonlin],
                self._ubg,
            )
        else:
            g = self._g
            lbg = self._lbg
            ubg = self._ubg

        f_k = self.f(self.x_sol_best, nlpdata.p)
        f_lin = self.grad_f_x_sub(self.x_sol_best, nlpdata.p)
        f_hess = self.f_hess(self.x_sol_best, nlpdata.p)
        f = f_k + f_lin.T @ dx + 0.5 * dx.T @ f_hess @ dx

        solver = ca.qpsol(
            f"benders_constraint_{self.nr_g}",
            self.settings.MIP_SOLVER,
            {"f": f, "g": g, "x": self._x, "p": self._nu},
            self.options,
        )

        sol = solver(
            x0=self.x_sol_best,
            lbx=nlpdata.lbx,
            ubx=nlpdata.ubx,
            lbg=lbg,
            ubg=ubg,
            p=[self.y_N_val - 1e-4],
        )

        nlpdata.solved, stats = self.collect_stats("BTR-MIP", solver, sol)
        if nlpdata.solved:
            nlpdata.prev_solution = sol
        else:
            nlpdata.prev_solution = self.best_data
            print("FINAL ITERATION")
            # Final iteration
            nlpdata.solved = True
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""A repository for MINLP solvers."""

from setuptools import setup, find_packages

setup(
    name="caminopy",
    version="0.1.4",
    description="Collection of Algorithms for Mixed-Integer Nonlinear Optimization",
    url="https://github.com/minlp-toolbox/CAMINO",
    author="Andrea Ghezzi, Wim Van Roy",
    author_email="andrea.ghezzi@imtek.uni-freiburg.de, wim.vr@hotmail.com",
    license="GPL-3.0",
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.8",
    install_requires=[
        "numpy >= 1.24.4",
        "pandas >= 2.0.3",
        "casadi >= 3.6.1",
        "scipy >= 1.10.1",
        "pytz >= 2024.2",
        "matplotlib >= 3.7.5",
        "parameterized >= 0.9.0",
        "timeout-decorator >= 0.5.0",
        "tox >= 4.1.2",
        "colored >= 1.4.4",
        "argcomplete >= 3.5.1",
        "seaborn >= 0.13.1",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from copy import deepcopy
from camino.data import MinlpData
from camino.solvers import MiSolverClass
from camino.solvers.sequential.milp_tr import TrustRegionMILP
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.utils import colored, toc

logger = logging.getLogger(__name__)


class SequentialTrustRegionMILP(MiSolverClass):

    def __init__(self, problem, data, stats, settings):
        super(SequentialTrustRegionMILP, self).__init__(problem, data, stats, settings)
        # Required subsolvers
        self.tr_milp = TrustRegionMILP(problem, data, stats, settings)
        self.nlp = NlpSolver(problem, stats, settings)
        # Algorithm parameters
        self.delta = 1.0  # Initial radius
        self.delta_max = 64  # Max delta
        self.eps = 1e-4  # Tolerance
        # Sufficient decrease, and update rhos:
        self.rho = [0.1, 0.1, 0.2]
        self.kappa = 0.5  # Reduction factor
        self.p = 0.5  # Monotonicity parameter

    def solve(self, nlpdata: MinlpData, with_nlp_improvement: bool = False):
        self.reset()
        nlpdata = self.nlp.solve(nlpdata, integers_relaxed=False)
        phi = nlpdata.obj_val
        logger.info(f"Initial start {phi=}")
        while True:
            data_p = self.tr_milp.solve(deepcopy(nlpdata), self.delta)
            logger.info(colored(f"TR-MILP {data_p.x_sol=}"))
            # Possible gain:
            psi_k = -data_p.obj_val
            # Make feasible if possible:
            if with_nlp_improvement:
                # This is a step, not described by De Marchi:
                data_p = self.nlp.solve(data_p, integers_relaxed=False)
                f = data_p.obj_val
            else:
                f = float(self.tr_milp.f(data_p.x_sol, data_p.p))
            if data_p.solved:
                logger.info(f"MILP-TR result {psi_k=}, {f=} < {phi=}")
                # Merit decrease
                a_k = phi - f
                logger.info(f"{a_k=} | {psi_k=} <? {self.eps=}")
                if psi_k < self.eps:
                    self.stats["total_time_calc"] = toc(reset=True)
                    colored("Problem solved", "green")
                    data_p.prev_solutions[0]["f"] = self.tr_milp.f(
                        data_p.x_sol, data_p.p
                    )
                    return data_p
                if a_k < self.rho[0] * psi_k:
                    self.delta = self.kappa * self.delta
                    colored(f"Trust region decreases to {self.delta}")
                else:
                    # Accept step
                    nlpdata = data_p
                    # Update merit function
                    phi = (1 - self.p) * phi + self.p * f
                    # Update step size
                    rho_k = a_k / psi_k
                    if rho_k < self.rho[1]:
                        self.delta = self.kappa * self.delta
                    elif rho_k < self.rho[2]:
                        self.delta = self.delta
                    else:
                        self.delta = self.delta / self.kappa
                    self.delta = min(self.delta, self.delta_max)
                    colored(f"Step accepted, trust radius {self.delta}", "green")
            else:
                raise NotImplementedError(
                    "The original paper doesn't handle nonlinear constraints, "
                    "this implementation can not make it feasible!"
                )

    def reset(self):
        """Reset."""
        self.delta = 1.0

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart."""
        logger.info(
            colored(
                "s-tr-milp warmstarting is automatic, new linearization point inherited from previous solver.",
                "yellow",
            )
        )



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

from camino.solvers import MiSolverClass
from camino.stats import Stats
from camino.settings import Settings
from camino.problem import MinlpProblem
from camino.data import MinlpData
from camino.solvers.decomposition import (
    GeneralizedBenders,
    GeneralizedBendersQP,
    OuterApproximation,
    OuterApproximationQP,
    OuterApproximationImproved,
    OuterApproximationQPImproved,
    SequentialVoronoiMIQP,
    SequentialBendersMIQP,
)
from camino.solvers.sequential import SequentialTrustRegionMILP
from camino.solvers.external.bonmin import BonminSolver
from camino.solvers.pumps import (
    FeasibilityPump,
    ObjectiveFeasibilityPump,
    RandomObjectiveFeasibilityPump,
)
from camino.solvers.approximation import CiaSolver
from camino.solvers.relaxation import NlpSolver
from camino.solvers.mip import MipSolver

SOLVER_MODES = {
    "gbd": GeneralizedBenders,
    "gbd-qp": GeneralizedBendersQP,
    "oa": OuterApproximation,
    "oa-qp": OuterApproximationQP,
    "oa-i": OuterApproximationImproved,
    "oa-qp-i": OuterApproximationQPImproved,
    "s-v-miqp": SequentialVoronoiMIQP,
    "s-b-miqp": SequentialBendersMIQP,
    "s-b-miqp-early-exit": lambda *args, **kwargs: SequentialBendersMIQP(
        *args, **kwargs, with_lb_milp=False
    ),
    "s-b-milp": lambda *args, **kwargs: SequentialBendersMIQP(
        *args, **kwargs, with_lb_milp=True, with_milp_only=True
    ),
    "s-tr-milp": SequentialTrustRegionMILP,
    "fp": FeasibilityPump,
    "ofp": ObjectiveFeasibilityPump,
    "rofp": RandomObjectiveFeasibilityPump,
    "bonmin": lambda *args, **kwargs: BonminSolver(*args, **kwargs, algo_type="B-BB"),
    "bonmin-bb": lambda *args, **kwargs: BonminSolver(
        *args, **kwargs, algo_type="B-BB"
    ),
    "bonmin-oa": lambda *args, **kwargs: BonminSolver(
        *args, **kwargs, algo_type="B-OA"
    ),
    "bonmin-qg": lambda *args, **kwargs: BonminSolver(
        *args, **kwargs, algo_type="B-QG"
    ),
    "bonmin-hyb": lambda *args, **kwargs: BonminSolver(
        *args, **kwargs, algo_type="B-Hyb"
    ),
    "bonmin-ifp": lambda *args, **kwargs: BonminSolver(
        *args, **kwargs, algo_type="B-iFP"
    ),
    "cia": CiaSolver,
    "nlp": lambda *args, **kwargs: NlpSolver(*args, **kwargs, integers_relaxed=True),
    "nlp-fxd": lambda *args, **kwargs: NlpSolver(*args, **kwargs, integers_relaxed=False),
    "mip": MipSolver,
}


def as_list(item, nr):
    """Make list if it is not a list yet."""
    if isinstance(item, list):
        return item
    else:
        return [item] * nr


class MinlpSolver(MiSolverClass):
    """Polysolver loading the required subsolver and solving the problem."""

    def __init__(
        self,
        name,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats = None,
        settings: Settings = None,
        problem_name: str = "generic",
        *args,
    ):
        if stats is None:
            stats = Stats(name, problem_name, "generic")
        if settings is None:
            settings = Settings()

        self.stats = stats
        self.settings = settings
        super(MinlpSolver, self).__init__(problem, data, stats, settings)

        # Create actual solver
        if name in SOLVER_MODES:
            self._subsolvers = [
                SOLVER_MODES[name](problem, data, stats, settings, *args)
            ]
            return
        elif "+" in name:
            names = name.split("+")
            for name in names:
                if name not in SOLVER_MODES:
                    raise Exception(f"Subsolver {name} does not exists")
            self._subsolvers = [
                SOLVER_MODES[subname](problem, data, stats, setting, *args)
                for subname, setting in zip(names, as_list(settings, len(names)))
            ]
        else:
            raise Exception(
                f"Solver mode {name} not implemented, options are:"
                + ", ".join(SOLVER_MODES.keys())
            )

    def solve(self, nlpdata: MinlpData, *args, **kwargs) -> MinlpData:
        """Solve the problem."""
        nlpdata = self._subsolvers[0].solve(nlpdata, *args, **kwargs)
        for subsolver in self._subsolvers[1:]:
            subsolver.warmstart(nlpdata)
            nlpdata = subsolver.solve(nlpdata, *args, **kwargs)

        return nlpdata

    def reset(self, nlpdata: MinlpData):
        """Reset the results."""
        [subsolver.reset() for subsolver in self._subsolvers]
        self.stats.reset()

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart."""
        self._subsolver[0].warmstart(nlpdata)

    def collect_stats(self):
        """Return the statistics."""
        return self.stats["success"], self.stats

    def get_settings(self):
        """Return settings."""
        return self.settings



# @Author: Massimo De Mauri <massimo>
# @Date:   2020-12-30T12:40:22+01:00
# @Email:  massimo.demauri@protonmail.com
# @Filename: models.py
# @Last modified by:   massimo
# @Last modified time: 2020-12-30T12:46:07+01:00
# @License: LGPL-3.0

import casadi as cs
from camino.settings import GlobalSettings

# engine


def insight_50kw_power_jc():

    # properties : jointly-convex
    w = GlobalSettings.CASADI_VAR.sym("w")
    P = GlobalSettings.CASADI_VAR.sym("P")

    wrat = GlobalSettings.CASADI_VAR.sym("wrat")
    Prat = GlobalSettings.CASADI_VAR.sym("Prat")

    Trat = 1.1349 * Prat / wrat

    # scaled inputs
    wn = w / wrat
    Pn = P / Prat

    # generic data
    dFmax = 0.1 * Prat / 1000
    # fuel_volume = 1.3353  # ml/g or l/kg
    # fuel_density = 0.7489  # Kg/l
    mass = (1.8 + 0.6) * Prat / 1000
    inertia = 0.1  # guessed
    minw = 0.1 * wrat  # guessed
    dFi = 1e-3 * Prat * 0.045665168578973 / 1.1

    # fuel consumption constraint
    # 1
    a00 = 0.198 - 0.5054 + 0.5006 + 0.0234
    a01 = 2.642 - 0.03629 - 1.438 - 0.09664 + 0.5312 - 0.3161
    a02 = 0.6187 + 0.0307 + 1.51 - 0.2206 + 1.585 - 1.498
    a10 = -1.01 - 0.02489
    a11 = -0.1102 + 0.005099 + 0.2856 + 0.2132 - 1.662 + 1.374
    a20 = 1.867 + 0.01006
    dFr1 = (
        Prat
        / 50e3
        * (a20 * wn**2 + a02 * Pn**2 + a11 * wn * Pn + a10 * wn + a01 * Pn + a00)
    )
    # 2
    a00 = 0.1662
    a01 = 1 - 1.438 - 0.09664 + 0.5312 - 0.3161 - 0.112 - 0.7949
    a02 = 0.1 + 1.51 - 0.2206 + 1.585 - 1.498 + 3.476 + 4.362
    a10 = 0.05455
    a11 = 0.4 + 0.2856 + 0.2132 - 1.662 + 1.374 - 3.029 - 3
    a20 = 1
    dFr2 = (
        Prat
        / 50e3
        * (a20 * wn**2 + a02 * Pn**2 + a11 * wn * Pn + a10 * wn + a01 * Pn + a00)
    )
    # 3
    a00 = -0.04825
    a01 = 2.274
    a02 = 0
    a10 = 0.6964
    a11 = 0
    a20 = 0
    dFr3 = (
        Prat
        / 50e3
        * (a20 * wn**2 + a02 * Pn**2 + a11 * wn * Pn + a10 * wn + a01 * Pn + a00)
    )
    # 4
    a00 = -2.507e-10
    a01 = 1.589
    a02 = 4 + 0.0246
    a10 = -2.867e-12
    a11 = -1
    a20 = 0.05 + 0.0246
    dFr4 = (
        Prat
        / 50e3
        * (a20 * wn**2 + a02 * Pn**2 + a11 * wn * Pn + a10 * wn + a01 * Pn + a00)
    )
    # 5
    a00 = -0.07
    a01 = 6
    a02 = 0
    a10 = -2.5
    a11 = 0
    a20 = 0
    dFr5 = (
        Prat
        / 50e3
        * (a20 * wn**2 + a02 * Pn**2 + a11 * wn * Pn + a10 * wn + a01 * Pn + a00)
    )
    # result
    dFr = cs.vertcat(dFr1, dFr2, dFr3, dFr4, dFr5)

    # max torque
    # 1
    p1 = 5.6e04
    p0 = 0
    Pmax1 = Prat / 50e3 * (p1 * wn + p0)
    # 2
    p1 = 2.68e04
    p0 = 2.334e04
    Pmax2 = Prat / 50e3 * (p1 * wn + p0)
    # result
    Pmax = cs.vertcat(Pmax1, Pmax2)

    return {
        "dFin": cs.Function("dFin", [wrat, Prat, w, P], [dFr]),
        "Fstart": cs.Function("Fstart", [Prat], [dFi]),
        "Pmax": cs.Function("Pmax", [wrat, Prat, w], [Pmax]),
        "Trat": cs.Function("Trat", [wrat, Prat], [Trat]),
        "mass": cs.Function("mass", [Prat], [mass]),
        "minw": cs.Function("minw", [wrat], [minw]),
        "inertia": cs.Function("inertia", [Prat], [inertia]),
        "dFmax": cs.Function("dFmax", [Prat], [dFmax]),
    }


# electric motor
def advisor_em_pwr_sc():
    # properties : semi-convex
    w = GlobalSettings.CASADI_VAR.sym("w")
    P = GlobalSettings.CASADI_VAR.sym("P")

    wrat = GlobalSettings.CASADI_VAR.sym("wrat")
    Prat = GlobalSettings.CASADI_VAR.sym("Prat")

    Trat = 3.7858 * Prat / wrat

    # scaled inputs
    wn = w / wrat
    Pn = P / Prat

    # generic data
    mass = 1.21 * Prat / 1000
    Pin_max = 1.2 * Prat
    inertia = 0.1

    c01 = -1.604e-12 * 0.85
    c02 = 0.1719 * 0.85
    c10 = 0.02417 * 0.85
    c20 = 2.895e-05 * 0.85
    Ploss0 = Prat * (c20 * wn**2 + c02 * Pn**2 + c10 * wn + c01 * Pn)

    c01 = 0
    c02 = 0.2425
    c10 = 0.6099
    c20 = -2.669
    Ploss1 = Prat * (c20 * wn**2 + c02 * Pn**2 + c10 * wn + c01 * Pn)

    c01 = -0.175
    c02 = 0.08767
    c10 = -0.3903
    c20 = 0.3944
    Ploss2 = Prat * (c20 * wn**2 + c02 * Pn**2 + c10 * wn + c01 * Pn)

    c01 = 0.175
    c02 = 0.08767
    c10 = -0.3903
    c20 = 0.3944
    Ploss3 = Prat * (c20 * wn**2 + c02 * Pn**2 + c10 * wn + c01 * Pn)

    Pin = cs.vertcat(P + Ploss0, P + Ploss1, P + Ploss2, P + Ploss3)
    # maps
    Pmax = cs.vertcat(Prat, Trat * wrat * wn)

    return {
        "Pin": cs.Function("Pin", [wrat, Prat, w, P], [Pin]),
        "Pmax": cs.Function("Pmax", [wrat, Prat, w], [Pmax]),
        "Trat": cs.Function("Trat", [wrat, Prat], [Trat]),
        "mass": cs.Function("mass", [Prat], [mass]),
        "inertia": cs.Function("inertia", [Prat], [inertia]),
        "Pin_max": cs.Function("Pin_max", [Prat], [Pin_max]),
    }


# battery
def battery_hu(Erat=GlobalSettings.CASADI_VAR.sym("Erat")):

    Q = 8280  # amp*s
    C = 51782  # farad
    R = 0.01  # ohm
    m = 0.07  # kilograms per cell
    imax = 70  # amp
    imin = -35  # amp
    U0 = 3.3  # volts

    SoC = GlobalSettings.CASADI_VAR.sym("Soc")  # state of charge
    P = GlobalSettings.CASADI_VAR.sym("P")  # internal power in watt

    n = Erat / (0.5 * Q**2 / C + Q * U0)  # num cells
    mass = m * n  # total battery mass
    E = Erat * SoC  # current energy

    dSoC = -P / Erat
    Pout = P - (R * C * P**2) / (2 * E + n * C * U0**2)
    # Pmax = imax*cs.sqrt(n*(2*Erat/C + n*U0**2))
    # Pmin = imin*cs.sqrt(n*(2*Erat/C + n*U0**2))
    Pmax = imax * cs.sqrt(n * (2 * E / C + n * U0**2))
    Pmin = imin * cs.sqrt(n * (2 * E / C + n * U0**2))

    return {
        "dSoC": cs.Function("dSoC", [Erat, SoC, P], [dSoC]),
        "Pout": cs.Function("Pout", [Erat, SoC, P], [Pout]),
        "Pmax": cs.Function("Pmax", [Erat, SoC], [Pmax]),
        "Pmin": cs.Function("Pmin", [Erat, SoC], [Pmin]),
        "SoC_max": 0.8,
        "SoC_min": 0.2,
        "mass": cs.Function("mass", [Erat], [mass]),
    }



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

# DISCLAIMER: this implementation is experimental and in general it does not work!

import casadi as ca
from camino.settings import Settings
from camino.solvers import SolverClass, Stats, MinlpProblem, MinlpData


class AmplSolver(SolverClass):
    """Create AMPL dump."""

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create MINLP problem."""
        super(AmplSolver, self).__init__(problem, stats, s)
        options = s.AMPL_EXPORT_SETTINGS.copy()
        options.update({"solver": "python3 -m minlp_algorithms copy /tmp/out.nl"})

        self.nr_x = problem.x.shape[0]
        discrete = [0] * self.nr_x
        for i in problem.idx_x_integer:
            discrete[i] = 1
        # options.update({
        #     "discrete": discrete,
        # })
        minlp = {
            "f": problem.f,
            "g": problem.g,
            "x": ca.vertcat(problem.x, problem.p),
            # "p": problem.p
        }
        self.solver = ca.nlpsol("ampl", "ampl", minlp, options)

    def solve(self, nlpdata: MinlpData, prev_feasible=True) -> MinlpData:
        """Solve MINLP."""
        try:
            nlpdata.prev_solution = self.solver(
                x0=nlpdata.x0,
                lbx=nlpdata.lbx,
                ubx=nlpdata.ubx,
                lbg=nlpdata.lbg,
                ubg=nlpdata.ubg,
                # p=nlpdata.p,
            )
        finally:
            print("File saved to /tmp/out.nl")

        raise Exception("See AMLP file!")
